;;;; delimited-control.scm
;;;; Kon Lovett, Apr 6 '06

;; prompt, prompt0, reset, reset0 are synonyms


;FIXME single-threaded

(module delimited-control

(;export
  (prompt $prompt) (control $control) (control-abort $control)
  (prompt0 $prompt) (control0 $control)
  (reset $prompt) (shift $prompt)
  (reset0 $control) (shift0 $control))

(import scheme)
(import (chicken base))
(import (chicken type))
(import (only miscmacros let/cc))

(: $prompt (procedure -> . *))
(: $control (boolean boolean procedure -> . *))

;;;

;; prompt-control

;

(define-syntax prompt
  (syntax-rules ()
    ((prompt EXPR) ($prompt (lambda () EXPR))) ) )

(define-syntax prompt0
  (syntax-rules ()
    ((prompt0 EXPR) ($prompt (lambda () EXPR))) ) )

; +F-

(define-syntax control
  (syntax-rules ()
    ((control PP EXPR) ($control #f #t (lambda (PP) EXPR))) ) )

(define-syntax control-abort
  (syntax-rules ()
    ((control-abort V) ($control #f #t (lambda (_) V))) ) )

; -F-

(define-syntax control0
  (syntax-rules ()
    ((control0 PP EXPR) ($control #f #f (lambda (PP) EXPR))) ) )

;; reset-shift

;

(define-syntax reset
  (syntax-rules ()
    ((reset EXPR) ($prompt (lambda () EXPR))) ) )

(define-syntax reset0
  (syntax-rules ()
    ((reset0 EXPR) ($prompt (lambda () EXPR))) ) )

; +F+

(define-syntax shift
  (syntax-rules ()
    ((shift PP EXPR) ($control #t #t (lambda (PP) EXPR))) ) )

; -F+

(define-syntax shift0
  (syntax-rules ()
    ((shift0 PP EXPR) ($control #t #f (lambda (PP) EXPR))) ) )

;srfi 1

;used once
(define-inline (last-pair ls)
  (let loop ((pr ls))
    (let ((nx (cdr pr)))
      (if (null? nx) pr
          (loop nx) ) ) ) )

;; This is one single global mutable cell, holding the stack of holes
;;
;; A hole has a continuation and a mark. The mark, #t for a delimiting hole. A
;; non-delimiting hole is just like the return from a regular function.

(define *holes* (make-parameter '()))

(define-inline (hole-push! hole) (*holes* (cons hole (*holes*))))

;used once
(define-inline (hole-list-push! holes-org)
  (when (pair? holes-org)
    (set-cdr! (last-pair holes-org) (*holes*))
    (*holes* holes-org) ) )

(define-inline (hole-pop!*)
  (let ((holes (*holes*)))
    (*holes* (cdr holes))
    (car holes) ) )

(define (error-control)
  (error 'control "missing toplevel delimited control operator") )

(define-inline (hole-pop!)
  (if (not (null? (*holes*))) (hole-pop!*)
      (error-control)) )

;;

(define-inline (cell-new v mark) (cons v mark))
(define-inline (cell-ref c) (car c))
(define-inline (cell-marked? c) (cdr c))
(define-inline (cell-unmark! c) (set-cdr! c #f))

;;

;used once
(define-inline (unwind-till-marked! keep-delimited?)
  (let loop ((hole (hole-pop!)) (holes-prefix '()))
    (if (cell-marked? hole)
        ;then it's prompt's hole, put it back & return the unmarked
        (begin
          (unless keep-delimited?
            ;make the hole non-delimiting
            (cell-unmark! hole)
            #; ;re-use?
            (cell-new (cell-ref hole) #f) )
          (hole-push! hole)
          holes-prefix )
        ;else save so can land on top of the, to be placed, mark
        (loop (hole-pop!) (cons hole holes-prefix)) ) ) )

; Essentially this is the ``return from the function''
(define-syntax *abort-top!
  (syntax-rules ()
    ((*abort-top! EXPR)
      (call-with-values
        (lambda () EXPR)
        (lambda vals (apply (cell-ref (hole-pop!*)) vals))) ) ) )

(define ($prompt thunk)
  (let/cc outer-k
    (hole-push! (cell-new outer-k #t)) ;it's prompt's hole
    (*abort-top! (thunk)) ) )

(define ($control shift? keep-delimited? func)
  (let/cc k-control
    (let* ((holes-prefix (unwind-till-marked! keep-delimited?))
           (invoke-subcont (lambda vals
                             (let/cc k-return
                               (hole-push! (cell-new k-return shift?))
                               (hole-list-push! holes-prefix)
                               (apply k-control vals)))) )
      (*abort-top! (func invoke-subcont)) ) ) )

) ;module delimited-control
